# IMPORT Packages and generate inputs for FASTSim (vehicles)

import pandas as pd
import numpy as np
import datetime
import matplotlib.pyplot as plt
from t3co import run_scenario
from t3co import generateinputs

generateinputs.generate()